"""Module to test pipeline telemetry helper methods."""

import test_data as td
from errors import ReturnValueWithErrorStatus, ReturnValueWithStatus

import pipeline_telemetry.helper as HP
from pipeline_telemetry import (
    Telemetry,
    add_errors_from_return_value,
    add_telemetry_counters_from_return_value,
    increase_base_count,
    increase_fail_count,
    is_telemetry_counter,
    process_return_value,
)
from pipeline_telemetry.settings.settings import (
    BASE_COUNT_KEY,
    COUNTERS_KEY,
    ERRORS_KEY,
    FAIL_COUNT_KEY,
)


class HelperTest:
    pass


def test_add_result_value_with_status_without_errors():
    """
    Test adding the ResultValueWithStatus instance without an error in it, to
    the telemetry adds no errors to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_without_error = ReturnValueWithStatus(result=[])
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_without_error,
    )
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert not getattr(telemetry_data, ERRORS_KEY)


def test_add_result_value_with_status_with_one_error():
    """
    Test adding the ResultValueWithStatus instance with an error in it to the
    telemetry adds the actual error to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_with_error = ReturnValueWithErrorStatus(error=td.TEST_ERROR_CODE)
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_with_error,
    )
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert td.TEST_ERROR_CODE.code in getattr(telemetry_data, ERRORS_KEY)
    assert getattr(telemetry_data, ERRORS_KEY).get(td.TEST_ERROR_CODE.code) == 1


def test_add_result_value_with_status_with_errors():
    """
    Test adding the ResultValueWithStatus instance with 2 errors in it to the
    telemetry adds the actual errors to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_with_error = ReturnValueWithErrorStatus(error=td.TEST_ERROR_CODE)
    # add the second error code to return_value_with_error
    return_value_with_error.add_error(td.TEST_ERROR_CODE)
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_with_error,
    )
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, ERRORS_KEY).get(td.TEST_ERROR_CODE.code) == 2


def test_process_telemetry_counters_in_return_value():
    """
    Test that when a ResultValueWithStatus instance with 2 telemetry counters
    (of which one with an error) in the result is processed by
    process_telemetry_counters_in_return_value both the error counter and
    counter are increased
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    add_telemetry_counters_from_return_value(
        object_with_telemetry=test_obj, return_value=td.TEST_RETURN_VALUE
    )
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, ERRORS_KEY) is not None
    assert getattr(telemetry_data, COUNTERS_KEY).get("test_counter") == 1


def test_process_telemetry_counters_in_return_value_returns_list():
    """
    Test that when a ResultValueWithStatus instance with 2 telemetry counters
    in the result is processed by process_telemetry_counters_in_return_value
    the result list without TelemetryCounters is returned.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    res_without_telem_counters = add_telemetry_counters_from_return_value(
        object_with_telemetry=test_obj, return_value=td.TEST_RETURN_VALUE
    )
    assert res_without_telem_counters == [
        res for res in td.TEST_RETURN_VALUE.result if not is_telemetry_counter(res)
    ]


def test_increase_base_count():
    """
    Test increase base count without increment increase base counter with 1 for
    a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    increase_base_count(object_with_telemetry=test_obj, sub_process="RETRIEVE_RAW_DATA")
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, BASE_COUNT_KEY) == 1


def test_increase_base_count_with_increment():
    """
    Test increase base count with increment 2 increases base counter with 2 for
    a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    increase_base_count(
        object_with_telemetry=test_obj, sub_process="RETRIEVE_RAW_DATA", increment=2
    )
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, BASE_COUNT_KEY) == 2


def test_increase_fail_count():
    """
    Test increase fail count without increment increases teh fail counter with
    1 for a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    test_obj._telemetry._initialize_sub_process("RETRIEVE_RAW_DATA")
    increase_fail_count(object_with_telemetry=test_obj, sub_process="RETRIEVE_RAW_DATA")
    telemetry_data = test_obj._telemetry.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, FAIL_COUNT_KEY) == 1


def test_process_return_value(mocker):
    """
    Test that process_return_value returns helper method returns the list
    retrieved add_telemetry_counters_from_return_value helper method.
    """
    mocker.patch(
        ("pipeline_telemetry.helper." "add_telemetry_counters_from_return_value"),
        return_value=["list with values"],
    )
    mocker.patch(
        ("pipeline_telemetry.helper." "add_errors_from_return_value"), return_value=None
    )
    assert process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE,
    ) == ["list with values"]


def test_process_return_value_processes_the_errors(mocker):
    """
    Test process_return_value calls the add_errors_from_return_value method.
    """
    mocker.patch(
        ("pipeline_telemetry.helper." "add_telemetry_counters_from_return_value"),
        return_value=None,
    )
    mocker.patch(
        ("pipeline_telemetry.helper." "add_errors_from_return_value"), return_value=None
    )
    process_telemetry_spy = mocker.spy(HP, "add_telemetry_counters_from_return_value")

    process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE,
    )
    assert process_telemetry_spy.called


def test_process_return_value_processes_the_telemetry_counters(mocker):
    """
    Test process_return_value calls the
    process_telemetry_counters_in_return_value method.
    """
    mocker.patch(
        ("pipeline_telemetry.helper." "add_telemetry_counters_from_return_value"),
        return_value=None,
    )
    mocker.patch(
        ("pipeline_telemetry.helper." "add_errors_from_return_value"), return_value=None
    )
    process_telemetry_spy = mocker.spy(HP, "add_telemetry_counters_from_return_value")

    process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE,
    )
    assert process_telemetry_spy.called



"""
Module to define a the TelemetryAggregator class.

This class creates makes an aggregation from a queryset of telemetry objects.
The aggregation of is added to a telemetry object that is provide to aggregate method of the class instance.

Usage

>>> # create new telemetry object
>>> telemetry_object = Telemetry(**telemtry_params)
>>> # create aggregator
>>> aggregator = TelemetryAggregator(telemetry_object)
>>> # aggregate a telemetry queryset and return the result
>>> aggregated_telemetry = aggregator.aggregate(telemetry_queryset)

"""

from abc import ABC
from datetime import date, datetime, timedelta
from typing import Iterator, Protocol, Type

from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings import settings as st
from pipeline_telemetry.settings.date_ranges import DateTimeRange, get_daily_date_ranges

from .helper import TelemetryAggregator, TelemetryListArgs, TelemetrySelector

DATE_TIME_RANGE_GENERATOR = {
    st.DAILY_AGGR_TELEMETRY_TYPE: get_daily_date_ranges,
}


class TelemetryStorage(Protocol):
    def telemetry_list(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator[TelemetryModel]:
        """
        Method to return an iteraror TelemetryModel instances retrieved from a
        database query with the provided arguments.
        """
        ...

    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        ...


class AbstractAggregator(ABC):
    """
    Aggregator to aggregate all SINGLE TELEMETRY objects for a single day
    into a telemetry objetc of type DAILY AGGREGATION.

    When initializig the class a TelemetrySelector should provided that
    determine which telemnetry objects are in scope

    The following methods are available to run the actual aggregations

    - aggregate: Method to implememt the aggregation
    """

    FROM_TELEMETRY_TYPE: str
    TO_TELEMETRY_TYPE: str

    __telemetry_selector: TelemetrySelector
    __target_telemetry: TelemetryModel
    __telemetry_storage: TelemetryStorage
    __aggregator: Type[TelemetryAggregator] = TelemetryAggregator

    def __init__(
        self, telemetry_selector: TelemetrySelector, telemetry_storage: TelemetryStorage
    ) -> None:
        self.__telemetry_selector = telemetry_selector
        self.__telemetry_storage = telemetry_storage
        self._set_target_telemetry_model()

    @property
    def target_telemetry(self):
        return self.__target_telemetry

    @property
    def storage_class(self):
        return self.__telemetry_storage

    def aggregate(self, start_date: date, end_date: date) -> None:
        """Method to run all aggregations in the given time period."""
        date_time_ranges = self._get_date_ranges(
            start_date=start_date, end_date=end_date
        )
        for date_time_range in date_time_ranges:
            aggregated_telemetry = self._run_aggregation(date_time_range)
            self.__telemetry_storage.store_telemetry(aggregated_telemetry)

    def _get_date_ranges(
        self, start_date: date, end_date: date
    ) -> Iterator[DateTimeRange]:
        """
        Method to return the list of date ranges based upon start and end
        date.
        """
        generator = DATE_TIME_RANGE_GENERATOR.get(self.TO_TELEMETRY_TYPE)
        if not generator:
            raise exceptions.UnknownTelemetryType(self.TO_TELEMETRY_TYPE)

        return generator(start_date=start_date, end_date=end_date)

    def _run_aggregation(self, date_time_range: DateTimeRange) -> TelemetryModel:
        """
        Method to run the actual aggregation and return the aggregated telemetry model.
        """
        # gather the database instances to be aggregated
        telemetry_list_params = self._telememtry_list_params(date_time_range)._asdict()
        telemetry_objects = self.__telemetry_storage.telemetry_list(
            **telemetry_list_params
        )

        initial_telemetry_obj = self.__target_telemetry.telemetry_copy()
        aggregator = self.__aggregator(initial_telemetry_obj)
        aggregated_telemetry = aggregator.aggregate(telemetry_objects)

        return self._set_start_date_time_for_aggregated_telemetry(
            aggregated_telemetry, date_time_range
        )

    @staticmethod
    def _set_start_date_time_for_aggregated_telemetry(
        telemetry_obj: TelemetryModel, date_time_range: DateTimeRange
    ) -> TelemetryModel:
        """
        Method to set the telemetry start_date_time attribute to the date_time
        for which the aggregation was run as start_date_time is set to now() by
        default.
        """
        # For daily telemetry aggregation the date of the aggregated object
        # is the from_date attribute in the date_time_range
        telemetry_obj.start_date_time = date_time_range.from_date
        return telemetry_obj

    def _set_target_telemetry_model(self) -> None:
        """
        Method to create and set the initial instance of the target telemetry
        model.
        """
        self.__target_telemetry = TelemetryModel(
            telemetry_type=self.TO_TELEMETRY_TYPE,
            source_name=self.__telemetry_selector.source_name,
            process_type=self.__telemetry_selector.process_type,
            category=self.__telemetry_selector.category,
            sub_category=self.__telemetry_selector.sub_category,
        )

    def _telememtry_list_params(
        self, date_time_range: DateTimeRange
    ) -> TelemetryListArgs:
        """
        Method to return telemetry_list_params that can be used to retrieve a
        list of TelemetryModel objects from the storage model.

        i.e. these params serve as input to the telemetry_list method of the
        storage_class
        """
        telemetry_list_params = self.__telemetry_selector._asdict()
        return TelemetryListArgs(
            telemetry_type=self.FROM_TELEMETRY_TYPE,
            from_date_time=date_time_range.from_date,
            to_date_time=date_time_range.to_date,
            **telemetry_list_params,
        )


class DailyAggregator(AbstractAggregator):
    """
    Aggregator to aggregate all SINGLE TELEMETRY objects for a single day
    into a telemetry objetc of type DAILY AGGREGATION.

    When initializig the class a TelemetrySelector should provided that
    determine which telemnetry objects are in scope

    The following methods are available to run the actual aggregations

    - aggregate: makes daily aggregations from start_date to end_date
    - aggregate_yesterday: makes daily aggregation for yesterday
    """

    FROM_TELEMETRY_TYPE = st.SINGLE_TELEMETRY_TYPE
    TO_TELEMETRY_TYPE = st.DAILY_AGGR_TELEMETRY_TYPE

    def aggregate_yesterday(self) -> None:
        """Method to run the aggregation for yesterday."""
        start_date = date.today()
        end_date = date.today() - timedelta(days=1)
        self.aggregate(start_date=start_date, end_date=end_date)


class PartialToSingleAggregator(AbstractAggregator):
    """
    Aggregator to aggregate all PARTIAL TELEMETRY objects for a single day
    into a telemetry objetc of type SINGLE AGGREGATION.

    When initializig the class a TelemetrySelector should provided that
    determine which telemnetry objects are in scope

    The following methods are available to run the actual aggregations

    - aggregate: makes daily aggregations from start_date to end_date
    """

    FROM_TELEMETRY_TYPE = st.PARTIAL_AGGR_TELEMETRY_TYPE
    TO_TELEMETRY_TYPE = st.SINGLE_TELEMETRY_TYPE



from .date_ranges import DateRange, DateTimeRange
from .settings import (
    AGGR_DATE_TIME_RANGE_METHODS,
    CATEGORY_KEY,
    DEFAULT_TRAFIC_LIGHT_COLOR,
    PROCESS_TYPE_KEY,
    SOURCE_NAME_KEY,
    START_TIME,
    SUB_CATEGORY_KEY,
    TELEMETRY_TYPE_KEY,
    TRAFIC_LIGHT_COLOR_ORANGE,
    TRAFIC_LIGHT_COLOR_RED,
)

__all__ = [
    "AGGR_DATE_TIME_RANGE_METHODS",
    "CATEGORY_KEY",
    "DEFAULT_TRAFIC_LIGHT_COLOR",
    "PROCESS_TYPE_KEY",
    "SOURCE_NAME_KEY",
    "START_TIME",
    "SUB_CATEGORY_KEY",
    "TELEMETRY_TYPE_KEY",
    "TRAFIC_LIGHT_COLOR_ORANGE",
    "TRAFIC_LIGHT_COLOR_RED",
    "DateRange",
    "DateTimeRange",
]



"""
Module to provide Telemetry class as public class to be accessed

classes
    - Telemetry

"""

from datetime import datetime
from typing import Dict, List, Optional, Type

from errors import ErrorCode

from .data_classes.telemetry_models import TelemetryData, TelemetryModel
from .helper import _raise_exception_if_telemetry_closed
from .settings import exceptions
from .settings import settings as st
from .settings.data_class import ProcessType, TelemetryCounter
from .settings.process_type import ProcessTypes
from .storage.generic import AbstractTelemetryStorage
from .storage.memory import TelemetryInMemoryStorage
from .validators.dict_validator import DictValidator


class Telemetry:
    _telemetry: TelemetryModel
    _telemetry_rules: dict
    _storage_class: Type[AbstractTelemetryStorage]
    _available_process_types: Type[ProcessTypes] = ProcessTypes
    _process_type: ProcessType
    _available_telemetry_types = st.TELEMETRY_TYPES

    def __init__(
        self,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: ProcessType,
        telemetry_type: str = st.DEFAULT_TELEMETRY_TYPE,
        telemetry_rules: Optional[dict] = None,
        storage_class: Type[AbstractTelemetryStorage] = TelemetryInMemoryStorage,
    ):
        self._process_type = process_type
        self._validate_process_type()
        self._storage_class = storage_class
        self._telemetry_rules = telemetry_rules or {}
        self._telemetry = TelemetryModel(
            telemetry_type=telemetry_type,
            category=category,
            sub_category=sub_category,
            source_name=source_name,
            process_type=process_type.name,
        )

    @classmethod
    def add_process_type(cls, process_type_key: str, process_type: ProcessType) -> None:
        """
        Add a custom process type to the available process types to the
        already registered process types.

        Args:
            process_type (ProcessType): Process type that needs to be added
            process_type_key (str): key that should be used when calling this
                                    ProcessType
        """
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType()
        cls._available_process_types.register_process_type(
            process_type_key, process_type
        )

    @property
    def storage_class(self) -> Type[AbstractTelemetryStorage]:
        """Storage_class property."""
        return self._storage_class

    @property
    def category(self) -> str:
        """Category property."""
        return getattr(self._telemetry, st.CATEGORY_KEY)

    @property
    def telemetry_type(self) -> str:
        """Telemetry_type property."""
        return getattr(self._telemetry, st.TELEMETRY_TYPE_KEY)

    @property
    def sub_category(self) -> str:
        """Category property."""
        return getattr(self._telemetry, st.SUB_CATEGORY_KEY)

    @property
    def source_name(self) -> str:
        """Source_name property."""
        return getattr(self.telemetry, st.SOURCE_NAME_KEY)

    @property
    def start_date_time(self) -> datetime:
        """start_date_time property."""
        return getattr(self.telemetry, st.START_TIME)

    @property
    def io_time_in_seconds(self) -> float:
        """io_time property."""
        return getattr(self.telemetry, st.IO_TIME_KEY)

    @property
    def run_time(self) -> float:
        """run_time property."""
        return getattr(self.telemetry, st.RUN_TIME)

    @property
    def traffic_light(self) -> str:
        """Traffic_light property."""
        return getattr(self.telemetry, st.TRAFFIC_LIGHT_KEY)

    @property
    def telemetry(self) -> TelemetryModel:
        """Telemetry property."""
        return self._telemetry

    @property
    def telemetry_data(self) -> Dict[str, TelemetryData]:
        """
        Telemetry_data property return only the telemetry datapoints
        datapoints from the telemetry object.
        """
        return getattr(self.telemetry, st.TELEMETRY_FIELD_KEY)

    @property
    def sub_process_types(self) -> list[str]:
        """Returns a of subprocess types allowed for the Telemetry instance."""
        return self._process_type.sub_processes

    def save_and_close(self) -> TelemetryModel:
        """
        Closes and stores the telemetry instance and returns the
        telemetry value.

        Returns:
            dict: telemetry result
        """
        if self.run_time:
            raise exceptions.TelemetryObjectAlreadyClosed()
        self._set_runtime()
        self._storage_class().store_telemetry(self.telemetry)

        return self.telemetry

    def add_telemetry_counter(
        self, telemetry_counter: TelemetryCounter, increment: Optional[int] = None
    ) -> None:
        """
        Method to process a TelemetryCounter object with predefined counters.

        Args:
            telemetry_counter (TelemetryCounter): [description]
            increment (int, None):
                when not None overules the increment setting from
                TelemetryCounter (which defaults to 1)
        """
        telemetry_counter.validate_sub_process()
        sub_process = telemetry_counter.sub_process
        counter_name = telemetry_counter.counter_name
        increment = increment or telemetry_counter.increment

        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        if telemetry_counter.error:
            self.increase_sub_process_error_count(
                sub_process=sub_process,
                error_code=telemetry_counter.error,
                increment=increment,
            )

        else:
            self.increase_sub_process_custom_count(
                sub_process=sub_process,
                custom_counter=counter_name,
                increment=increment,
            )

    @_raise_exception_if_telemetry_closed
    def add(self, sub_process: str, data: dict, errors: List[ErrorCode]) -> None:
        """
        Add data validation errors and/or errors from data process to a
        telemetry sub process.
        Data validation errors are retrieved from data validation defined in
        telemetry rules.

        Method does not update BASE_COUNT.

        :param sub_process: applicable subprocess for the errors
        :type sub_process: str
        :param data: data generated by subproces for which telemetry
                     needs to be generated
        :param type: dict
        :param errors: errors from data process to be added
        :type errors: list of Errorcodes
        :returns: None
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)
        validation_errors = self._validate_data(sub_process, data)
        # add data validation errors
        self._add_errors(sub_process, validation_errors)

        # add data process errors
        if errors:
            self._add_errors(sub_process, errors)

    def _validate_data(self, sub_process: str, data: dict) -> List[ErrorCode]:
        """Validates the data provided by a subprocess.

        Data is validated based upon telemetry rules for the subprocces

        :param sub_process: applicable subprocess
        :typesub_process: str
        :param data: data provided by subprocess
        :type data: dict
        :returns: list of Errorcodes
        """
        telemetry_rules = self._telemetry_rules.get(sub_process, {})
        return DictValidator().validate(data, telemetry_rules)

    def _add_errors(self, sub_process: str, errors: list[ErrorCode]) -> None:
        """Adds the error to a telemetry sub process.

        :param sub_process: applicable subprocess for the errors
        :typesub_process: str
        :param errors: errors to be processed
        :type errors: list of Errorcodes
        :returns: None
        """
        for error_code in errors:
            self.increase_sub_process_error_count(
                error_code=error_code, sub_process=sub_process
            )

    @_raise_exception_if_telemetry_closed
    def set_orange_traffic_light(self) -> None:
        """Sets traffic light attribute to orange."""
        self.telemetry.set_orange_traffic_light()

    @_raise_exception_if_telemetry_closed
    def set_red_traffic_light(self) -> None:
        """Sets traffic light attribute to red."""
        self.telemetry.set_red_traffic_light()

    def _set_runtime(self) -> None:
        """Method to calculate and set the runtime seconds (in str)."""
        run_time = datetime.now() - self.start_date_time
        setattr(self.telemetry, st.RUN_TIME, round(run_time.total_seconds(), 2))

    def get(self, sub_process: str) -> TelemetryData:
        """Returns the data object for a sub process.

        Args:
            sub_process (str): name of sub_process name

        Raises:
            exceptions.BaseCountForSubProcessNotAdded:
                when subprocess is not yet initialized

        Returns:
            TelemetryData: the sub_process TelemetryData dataclass object
        """
        if self._sub_process_not_yet_initialized(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        return self.telemetry_data[sub_process]

    @_raise_exception_if_telemetry_closed
    def increase_io_time(self, incremental_io_time: float) -> None:
        """
        Increases the io_time of the telemetry object.

        Args:
            io_time (float): io_time that needs to be added to the total io_time
        """
        current_io_time = getattr(self._telemetry, st.IO_TIME_KEY)
        increased_io_time = current_io_time + incremental_io_time
        setattr(self._telemetry, st.IO_TIME_KEY, increased_io_time)

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_base_count(
        self, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases the base count for a subprocess.

        Args:
            sub_process (str): name of subprocess
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        self.get(sub_process).increase_base_count(increment)

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_fail_count(
        self, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases the fail count for a subprocess.

        Args:
            sub_process (str): name of subprocess
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        self.get(sub_process).increase_fail_count(increment)

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_error_count(
        self, sub_process: str, error_code: ErrorCode, increment: int = 1
    ) -> None:
        """
        Increases a error counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            error_code (ErrorCode): ErrorCode object
            increment (int): increment for the counter

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if self._sub_process_not_yet_initialized(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self.get(sub_process).increase_error_count(
            increment=increment, error_code=error_code
        )

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_custom_count(
        self, custom_counter: str, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases a custom counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            custom_counter (str): name of custom counter
            increment (int): increment for the counter (default = 1)

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        self.get(sub_process).increase_custom_count(
            increment=increment, counter=custom_counter
        )

    def _sub_process_is_initialized(self, sub_process: str) -> bool:
        """Returns True if provided sub_process is initialized

        Args:
            sub_process (str): sub_process

        Returns:
            bool: True if sub_process is inialized, False otherwise
        """
        return sub_process in self.telemetry_data

    def _sub_process_not_yet_initialized(self, sub_process: str) -> bool:
        """Returns True if provided sub_process is not yet initialized

        Args:
            sub_process (str): sub_process

        Returns:
            bool: True if sub_process is inialized, False otherwis
        """
        return not self._sub_process_is_initialized(sub_process)

    def _validate_process_type(self) -> None:
        """Validates the process type for the telemetrty instance."""
        if not isinstance(self._process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType
        if not self._available_process_types.is_registered(self._process_type):
            raise exceptions.ProcessTypeNotRegistered(self._process_type)

    def _initialize_sub_process(self, sub_process: str) -> None:
        """Initializes a sun process in the telemetry data.

        Args:
            sub_process (str): sub process name
        """
        if sub_process not in self.sub_process_types:
            raise exceptions.InvalidSubProcess(sub_process, self._process_type)

        if sub_process in self.telemetry_data:
            raise exceptions.SubProcessAlreadyInitialized(sub_process)

        self.telemetry_data.update(**{sub_process: TelemetryData()})



"""Module to test storage module."""

import sqlite3
from datetime import datetime, timedelta

import pytest
from freezegun import freeze_time
from test_storage_data import DEFAULT_TELEMETRY_MODEL_PARAMS, DEFAULT_TELEMETRY_PARAMS

from pipeline_telemetry import Telemetry
from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings import DateTimeRange, exceptions
from pipeline_telemetry.settings import settings as st
from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage


def telemetry_query_params():
    """Returns a default set of query params"""
    return DEFAULT_TELEMETRY_MODEL_PARAMS | {
        "from_date_time": datetime.now() - timedelta(days=1),
        "to_date_time": datetime.now() + timedelta(days=1),
    }


def test_in_abstract_strorage_class_exists():
    """Test that abstract storage class exists."""
    assert AbstractTelemetryStorage


def test_in_memory_strorage_class_exists():
    """Test that in memory storage class exists."""
    assert TelemetryInMemoryStorage


def test_db_object_to_dict_method():
    """
    Test _db_object_to_dict method exists and if no override is done returns an
    object unchanged.
    """
    in_memory_storage = TelemetryInMemoryStorage()
    assert in_memory_storage._db_object_to_dict({"test": "dict"}) == {"test": "dict"}


def test_get_aggr_telem_query_params():
    """
    Test _get_aggr_telem_query_params converts TelemetryModel into
    dict with the correct keys.
    """
    telemetry_params = DEFAULT_TELEMETRY_MODEL_PARAMS.copy() | {
        "telemetry_type": "DAILY AGGREGATION"
    }

    telemetry = TelemetryModel(**telemetry_params)
    in_memory_storage = TelemetryInMemoryStorage()
    aggr_telem_query_params = in_memory_storage._get_aggr_telem_query_params(telemetry)
    assert isinstance(aggr_telem_query_params, dict)
    keys = [
        "telemetry_type",
        "source_name",
        "category",
        "sub_category",
        "process_type",
        "from_date_time",
        "to_date_time",
    ]
    assert all([key in aggr_telem_query_params.keys() for key in keys])
    assert not [key for key in aggr_telem_query_params.keys() if key not in keys]


def test_in_memory_storage_creates_db_in_memory():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    in_memory_storage = TelemetryInMemoryStorage()
    assert in_memory_storage.db_in_memory
    assert in_memory_storage.db_cursor
    assert isinstance(in_memory_storage.db_in_memory, sqlite3.Connection)
    assert isinstance(in_memory_storage.db_cursor, sqlite3.Cursor)


def test_store_telemetry_raise_exception_when_db_is_closed():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.close_db()
    with pytest.raises(exceptions.StorageNotInitialized):
        in_memory_storage.store_telemetry(
            TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
        )


def test_select_records_raise_exception_when_db_is_closed():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.close_db()
    with pytest.raises(exceptions.StorageNotInitialized):
        in_memory_storage.select_records(**telemetry_query_params())


def test_close_db_in_memory():
    """Test in memory storage close method close the db."""
    in_memory_storage = TelemetryInMemoryStorage()
    db_in_memory = in_memory_storage.db_in_memory
    cursor = in_memory_storage.db_cursor
    assert db_in_memory
    assert cursor
    in_memory_storage.close_db()
    with pytest.raises(sqlite3.ProgrammingError):
        cursor.execute("SELECT * FROM telemetry ")
    with pytest.raises(sqlite3.ProgrammingError):
        db_in_memory.cursor()


def test_close_db_in_memory_can_be_closed_twice():
    """
    Test in memory storage close method can be called
    when database is already closed.
    """
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.close_db()
    assert in_memory_storage.close_db() is None


def test_in_memory_strorage_is_only_created_once():
    """Test that once the database is created it will not be reinitialized"""
    in_memory_storage = TelemetryInMemoryStorage()
    db = in_memory_storage.db_in_memory
    TelemetryInMemoryStorage.initialize_db()
    assert in_memory_storage.db_in_memory is db
    new_in_memory_storage = TelemetryInMemoryStorage()
    assert new_in_memory_storage.db_in_memory is db


def test_store_telemetry_stores_object():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    all_in_memory_objects = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry "
    )
    assert len(all_in_memory_objects.fetchall()) == 1


def test_select_records_returns_sql_lite_cursor():
    """Test select_records method returns sqllite Cursor with one object."""
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    selected_in_memory_objects = in_memory_storage.select_records(
        **telemetry_query_params()
    )
    assert len(selected_in_memory_objects.fetchall()) == 1
    assert isinstance(selected_in_memory_objects, sqlite3.Cursor)


def test_records_with_outside_date_range_are_not_returned():
    """
    Test select_records method returns only those records within the given date
    range.
    """
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_storage.store_telemetry(
        TelemetryModel(
            **DEFAULT_TELEMETRY_MODEL_PARAMS
            | {"start_date_time": datetime.now() - timedelta(days=2)}
        )
    )
    selected_in_memory_objects = in_memory_storage.select_records(
        **telemetry_query_params()
    )
    assert len(selected_in_memory_objects.fetchall()) == 1
    all_in_memory_objects = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry "
    )
    assert len(all_in_memory_objects.fetchall()) == 2


def test_select_records_returns_only_selected_records():
    """
    Test select_records method returns only records accordding to giv.encode()attriubures.
    """
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_storage.store_telemetry(
        TelemetryModel(
            **DEFAULT_TELEMETRY_MODEL_PARAMS | {"category": "different_category"}
        )
    )
    selected_in_memory_objects = in_memory_storage.select_records(
        **telemetry_query_params()
    )
    assert len(selected_in_memory_objects.fetchall()) == 2
    all_in_memory_objects = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry "
    )
    assert len(all_in_memory_objects.fetchall()) == 3


@freeze_time("2022-01-18 18:00:00.123456")
def test_store_telemetry_stores_object_with_create_date(mocker):
    """Test in memory storage instance has db_in_memory and db_cursor"""
    frozen_time = {"start_date_time": datetime.now()}
    telemetry = DEFAULT_TELEMETRY_MODEL_PARAMS | frozen_time
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**telemetry))
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    assert in_memory_record["start_date_time"] == "2022-01-18T18:00:00.123456"


def test_telemetry_storage_to_object_returns_telemetry_model():
    """
    Test _telemetry_storage_to_object method returns a TelemetryModel instance
    created from an in memory storage object.
    """
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    telemetry = in_memory_storage._telemetry_storage_to_object(in_memory_record)
    assert isinstance(telemetry, TelemetryModel)


def test_telemetry_storage_to_object_returns_object_with_correct_values():
    """
    Test _telemetry_storage_to_object method returns a TelemetryModel instance
    with the correct values.
    """
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    telemetry = in_memory_storage._telemetry_storage_to_object(in_memory_record)
    for key, value in DEFAULT_TELEMETRY_MODEL_PARAMS.items():
        assert getattr(telemetry, key) == value


def test_telemetry_storage_to_object_returns_object_with_telemetry_data():
    """
    Test _telemetry_storage_to_object method returns a TelemetryModel instance
    with the correct values.
    """
    counter_key = "TEST_COUNTER"
    telemetry = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    sub_process = telemetry.get_sub_process_data("sub_process")
    sub_process.increase_base_count(increment=1)
    sub_process.increase_custom_count(increment=2, counter=counter_key)
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(telemetry)
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    telemetry_model = in_memory_storage._telemetry_storage_to_object(in_memory_record)
    telemetry_data = telemetry_model.get_sub_process_data("sub_process")
    assert telemetry_data
    assert getattr(telemetry_data, st.BASE_COUNT_KEY) == 1
    assert getattr(telemetry_data, st.COUNTERS_KEY)[counter_key] == 2


@freeze_time("2022-01-18 18:00:00.123456")
def test_get_aggr_telem_date_time_range():
    """
    Test _get_aggr_telem_date_time_range returns the proper date time range.
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy() | {
        "telemetry_type": "DAILY AGGREGATION"
    }
    telemetry = Telemetry(**telemetry_params)
    telemetry._telemetry.start_date_time = datetime.now()
    telemetry.start_date_time
    aggr_telem_date_time_range = (
        TelemetryInMemoryStorage()._get_aggr_telem_date_time_range(telemetry)
    )

    assert isinstance(aggr_telem_date_time_range, DateTimeRange)
    assert aggr_telem_date_time_range.from_date == datetime(2022, 1, 18)
    assert aggr_telem_date_time_range.to_date == datetime(2022, 1, 19)


@freeze_time("2022-01-18 18:00:00.123456")
def test_get_aggr_telem_date_time_range_monthly_aggr():
    """
    Test _get_aggr_telem_date_time_range returns the proper date time range
    with monthly telemetry aggregation type
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy() | {
        "telemetry_type": "MONTHLY AGGREGATION"
    }
    telemetry = Telemetry(**telemetry_params)
    telemetry._telemetry.start_date_time = datetime.now()
    telemetry.start_date_time
    aggr_telem_date_time_range = (
        TelemetryInMemoryStorage()._get_aggr_telem_date_time_range(telemetry)
    )

    assert aggr_telem_date_time_range.from_date == datetime(2022, 1, 1)
    assert aggr_telem_date_time_range.to_date == datetime(2022, 2, 1)


@freeze_time("2022-01-18 18:00:00.123456")
def test_get_aggr_telem_date_time_range_weekly_aggr():
    """
    Test _get_aggr_telem_date_time_range returns the proper date time range
    with weekly telemetry aggregation type
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy() | {
        "telemetry_type": "WEEKLY AGGREGATION"
    }
    telemetry = Telemetry(**telemetry_params)
    telemetry._telemetry.start_date_time = datetime.now()
    telemetry.start_date_time
    aggr_telem_date_time_range = (
        TelemetryInMemoryStorage()._get_aggr_telem_date_time_range(telemetry)
    )

    assert aggr_telem_date_time_range.from_date == datetime(2022, 1, 17)
    assert aggr_telem_date_time_range.to_date == datetime(2022, 1, 24)


def test_delete_telemetry_deletes_telemetry_record():
    """
    Test _telemetry_storage_to_object method returns a TelemetryModel instance
    with the correct values.
    """
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS))
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    telemetry = in_memory_storage._telemetry_storage_to_object(in_memory_record)
    for key, value in DEFAULT_TELEMETRY_MODEL_PARAMS.items():
        assert getattr(telemetry, key) == value



"""
Module to test telemetry model class for pipeline telemetry module.
"""

import pytest
from test_data import DEFAULT_TELEMETRY_MODEL_PARAMS

from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings import settings as st

# pylint: disable=protected-access


def test_telemetry_model_exists():
    """check that TelemetryModel class exists"""
    assert TelemetryModel


def test_check_telemetry_type_raises_exception_with_inv_type():
    telemetry_params = DEFAULT_TELEMETRY_MODEL_PARAMS.copy()
    telemetry_params[st.TELEMETRY_TYPE_KEY] = "invalid"
    with pytest.raises(exceptions.InvalidTelemetryType):
        TelemetryModel(**telemetry_params)._check_telemetry_type()


def test_telemetry_model_can_be_added():
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_added = tel_model_1 + tel_model_2
    assert isinstance(tel_model_added, TelemetryModel)


def test_telemetry_model_addition_adds_aggregation_counter():
    """
    Test that adding up to telemetry model instances return a Telemetry
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_added = tel_model_1 + tel_model_2
    aggregation_data = tel_model_added.telemetry[st.AGGREGATION_KEY]
    assert getattr(aggregation_data, st.BASE_COUNT_KEY) == 1


def test_telemetry_model_addition_aggregates_custom_counters():
    """
    Test that adding up to telemetry model instances returns a TelemetryModel
    instance with aggregated Traffic Ligth counters in aggregation telemetry
    subprocess.
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_3 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2.get_sub_process_data("sub_process").increase_custom_count(
        increment=1, counter="counter_1"
    )
    tel_model_2.get_sub_process_data("sub_process").increase_custom_count(
        increment=2, counter="counter_2"
    )
    tel_model_3.get_sub_process_data("sub_process").increase_custom_count(
        increment=2, counter="counter_1"
    )
    tel_model_3.get_sub_process_data("sub_process").increase_custom_count(
        increment=4, counter="counter_3"
    )
    tel_model_added = tel_model_1 + tel_model_2
    tel_model_added += tel_model_3
    aggregation_data = tel_model_added.telemetry[st.AGGREGATION_KEY]
    telemetry_data = tel_model_added.telemetry["sub_process"]
    assert getattr(aggregation_data, st.COUNTERS_KEY)[st.TRAFIC_LIGHT_COLOR_GREEN] == 2
    assert getattr(telemetry_data, st.COUNTERS_KEY) == {
        "counter_1": 3,
        "counter_2": 2,
        "counter_3": 4,
    }


def test_telemetry_model_addition_aggregates_errors():
    """
    Test that adding up to telemetry model instances returns a TelemetryModel
    instance with aggregated errors.
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_3 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2.get_sub_process_data("sub_process")._increase_error_count(
        increment=1, error_code_key="error_1"
    )
    tel_model_2.get_sub_process_data("sub_process")._increase_error_count(
        increment=2, error_code_key="error_2"
    )
    tel_model_3.get_sub_process_data("sub_process")._increase_error_count(
        increment=2, error_code_key="error_1"
    )
    tel_model_3.get_sub_process_data("sub_process")._increase_error_count(
        increment=4, error_code_key="error_3"
    )
    tel_model_added = tel_model_1 + tel_model_2
    tel_model_added += tel_model_3
    telemetry_data = tel_model_added.telemetry["sub_process"]
    assert getattr(telemetry_data, st.ERRORS_KEY) == {
        "error_1": 3,
        "error_2": 2,
        "error_3": 4,
    }


def test_telemetry_model_addition_aggregates_traffic_lights():
    """
    Test that adding up to telemetry model instances returns a TelemetryModel
    instance with aggregated Traffic Ligth counters in aggregation telemetry
    subprocess.
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(
        **(
            DEFAULT_TELEMETRY_MODEL_PARAMS
            | {"traffic_light": st.TRAFIC_LIGHT_COLOR_ORANGE}
        )
    )
    tel_model_3 = TelemetryModel(
        **(
            DEFAULT_TELEMETRY_MODEL_PARAMS
            | {"traffic_light": st.TRAFIC_LIGHT_COLOR_RED}
        )
    )
    tel_model_added = tel_model_1 + tel_model_2
    tel_model_added += tel_model_3
    aggregation_data = tel_model_added.telemetry[st.AGGREGATION_KEY]
    aggregation_data_counters = getattr(aggregation_data, st.COUNTERS_KEY)
    assert st.TRAFIC_LIGHT_COLOR_ORANGE in aggregation_data_counters
    assert st.TRAFIC_LIGHT_COLOR_RED in aggregation_data_counters
    assert aggregation_data_counters[st.TRAFIC_LIGHT_COLOR_RED] == 1
    assert aggregation_data_counters[st.TRAFIC_LIGHT_COLOR_ORANGE] == 1


def test_telemetry_model_addition_aggregates_io_time():
    """
    Test that adding up to telemetry model instances returns a TelemetryModel
    instance with aggregated io time rounded in seconds
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(
        **(DEFAULT_TELEMETRY_MODEL_PARAMS | {"io_time_in_seconds": 1.2})
    )
    tel_model_3 = TelemetryModel(
        **(DEFAULT_TELEMETRY_MODEL_PARAMS | {"io_time_in_seconds": 1.7})
    )
    tel_model_added = tel_model_1 + tel_model_2
    tel_model_added += tel_model_3
    aggregation_data = tel_model_added.telemetry[st.AGGREGATION_KEY]
    aggregation_data_counters = getattr(aggregation_data, st.COUNTERS_KEY)
    assert "io_time_in_seconds" in aggregation_data_counters
    assert aggregation_data_counters["io_time_in_seconds"] == 3


def test_telemetry_model_addition_aggregates_run_time():
    """
    Test that adding up to telemetry model instances returns a TelemetryModel
    instance with aggregated run time rounded in seconds
    """
    tel_model_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
    tel_model_2 = TelemetryModel(
        **(DEFAULT_TELEMETRY_MODEL_PARAMS | {"run_time_in_seconds": 1.2})
    )
    tel_model_3 = TelemetryModel(
        **(DEFAULT_TELEMETRY_MODEL_PARAMS | {"run_time_in_seconds": 1.7})
    )
    tel_model_added = tel_model_1 + tel_model_2
    tel_model_added += tel_model_3
    aggregation_data = tel_model_added.telemetry[st.AGGREGATION_KEY]
    aggregation_data_counters = getattr(aggregation_data, st.COUNTERS_KEY)
    assert "run_time_in_seconds" in aggregation_data_counters
    assert aggregation_data_counters["run_time_in_seconds"] == 3



"""
module to test entries_have_key instruction class
"""

from pipeline_telemetry.validators.entries_have_key import (
    EntriesHaveKey,
    EntriesHaveKeyRuleData,
)

# pylint: disable=protected-access


def test_has_key_class_exists():
    """check that HasKey class exists"""
    assert EntriesHaveKey


def test_has_key_class_is_singelton():
    """check that HasKey is a singleton class"""
    instance_one = EntriesHaveKey()
    instance_two = EntriesHaveKey()
    assert instance_one is instance_two


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "test", "must_have_key": "test"}
    )
    assert EntriesHaveKey._get_field_name(rule_content) == "test"


def test_get_must_have_key_method():
    """
    check that _get_must_have_key method returns value of
    must_have_key that is in scope of validation
    """
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "fieldname", "must_have_key": "test"}
    )
    assert EntriesHaveKey._get_must_have_key(rule_content) == "test"


def test_validate_field_exists():
    """
    test that _validate_field_exists method returns empty list when
    entries field exists
    """
    dict_to_validate = {"items": [{"key": 1}, {"key": 2}]}
    rule_data = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    assert EntriesHaveKey._validate_field_exists(dict_to_validate, rule_data) == []


def test_validate_field_exists_returns_error():
    """
    test that _validate_field_exists method returns empty list when
    entries field exists
    """
    dict_to_validate = {"not_items": [{"key": 1}, {"key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )
    errors = EntriesHaveKey._validate_field_exists(dict_to_validate, rule_content)

    assert len(errors) == 1
    assert "Entries field does not exist" in errors[0].description


def test_validate_entries_have_key():
    """
    test that _validate_entries_have_key method returns empty list when all
    entries have the right key
    """
    dict_to_validate = {"items": [{"key": 1}, {"key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    assert (
        EntriesHaveKey._validate_entries_have_key(dict_to_validate, rule_content) == []
    )


def test_validate_entries_have_nested_key():
    """
    test that _validate_entries_have_key method returns empty list when all
    entries have the right key
    """
    dict_to_validate = {"items": [{"key": {"nested_key": 2}}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key.nested_key"}
    )

    assert (
        EntriesHaveKey._validate_entries_have_key(dict_to_validate, rule_content) == []
    )


def test_validate_entries_have_key_with_missing_key():
    """
    test that _validate_entries_have_key method returns list with has_key
    error when a key is missing
    """
    dict_to_validate = {"items": [{"not_key": 1}, {"key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    errors = EntriesHaveKey._validate_entries_have_key(dict_to_validate, rule_content)

    assert len(errors) == 1
    assert "Key missing in entry" in errors[0].description


def test_validate_entries_have_key_with_non_dict_items():
    """
    test that _validate_entries_have_key method returns list with
    entry_is_not_a_dict error when entry is not a dicy
    """
    dict_to_validate = {"items": ["not_a_dict", {"key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    errors = EntriesHaveKey._validate_entries_have_key(dict_to_validate, rule_content)

    assert len(errors) == 1
    assert "Entry that needs to have a key is not a dict" in errors[0].description


def test_validate_entries_have_keys_with_missing_key():
    """
    test that _validate_entries_have_key method returns list with has_key
    errors when a all keys are missing
    """
    dict_to_validate = {"items": [{"not_key": 1}, {"also_not_key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    errors = EntriesHaveKey._validate_entries_have_key(dict_to_validate, rule_content)

    assert len(errors) == 2


def test_validate_calls_validate_field_exists(mocker):
    """
    test that _validate method calls the _validate_field_exists method
    """
    dict_to_validate = {"not_items": [{"not_key": 1}, {"also_not_key": 2}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    _validate_field_exists_spy = mocker.spy(EntriesHaveKey, "_validate_field_exists")

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_field_exists_spy.called


def test_validate_calls_validate_type_entries_field(mocker):
    """
    test that _validate method calls the _validate_type_entries_field method
    """
    dict_to_validate = {"items": "not a list or a dict"}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    _validate_type_entries_field_spy = mocker.spy(
        EntriesHaveKey, "_validate_type_entries_field"
    )

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_type_entries_field_spy.called


def test_validate_calls_validate_entries_have_key(mocker):
    """
    test that _validate method calls the _validate_type_entries_field method
    """
    dict_to_validate = {"items": [{"a": 1}]}
    rule_content = EntriesHaveKeyRuleData(
        **{"field_name": "items", "must_have_key": "key"}
    )

    _validate_entries_have_key_spy = mocker.spy(
        EntriesHaveKey, "_validate_type_entries_field"
    )

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_entries_have_key_spy.called


def test_validate_calls_validate(mocker):
    """
    test that public validate method calls the _validate method
    """
    dict_to_validate = {"items": [{"a": 1}]}
    rule_content = {"field_name": "items", "must_have_key": "key"}

    _validate_spy = mocker.spy(EntriesHaveKey, "_validate")
    EntriesHaveKey.validate(dict_to_validate, rule_content)

    assert _validate_spy.called



"""[summary]"""

import json
import sqlite3
from datetime import datetime
from typing import Iterator, Optional

from ..data_classes import TelemetryModel
from ..settings import exceptions
from ..settings import settings as st
from .generic import AbstractTelemetryStorage


def dict_factory(cursor, row):
    """Method to allow sql queries to be returned as a dict.

    The telemetry field is seperately converted to a dict because
    sqlite returns a nested dict as a string.
    """
    col_names = [col[0] for col in cursor.description]
    telemetry = {key: value for key, value in zip(col_names, row)}
    telemetry_str = telemetry.pop("telemetry", "{}")
    telemetry.update({"telemetry": json.loads(telemetry_str)})
    return telemetry


class TelemetryInMemoryStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in memory storage for use when unit testing
    """

    db_in_memory: Optional[sqlite3.Connection] = None
    db_cursor: Optional[sqlite3.Cursor] = None

    def __init__(self):
        if not self.db_in_memory:
            self.initialize_db()

    @classmethod
    def initialize_db(cls) -> None:
        """
        class method to initialize the in memory db
        """
        if not cls.db_in_memory:
            cls.db_in_memory = sqlite3.connect(":memory:")
            cls.db_in_memory.row_factory = dict_factory
            cls.db_cursor = cls.db_in_memory.cursor()
            cls._define_db_table(cls.db_cursor)

    @classmethod
    def close_db(cls) -> None:
        """close cursor and connections"""
        if cls.db_cursor:
            cls.db_cursor.close()

        if cls.db_in_memory:
            cls.db_in_memory.close()

        cls.db_cursor = None
        cls.db_in_memory = None

    @staticmethod
    def _define_db_table(cursor):
        """define telemetry table"""
        cursor.executescript(
            """
            DROP TABLE IF EXISTS telemetry;
            CREATE TABLE telemetry (telemetry_type varchar(100),
            category varchar(60), sub_category varchar(60),
            source_name varchar(40), process_type varchar(40),
            start_date_time timestamp, run_time_in_seconds varchar(20),
            telemetry json, traffic_light varchar(10),
            io_time_in_seconds real)"""
        )

    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        if not self.db_cursor:
            raise exceptions.StorageNotInitialized

        telemetry_type = getattr(telemetry, st.TELEMETRY_TYPE_KEY)
        category = getattr(telemetry, st.CATEGORY_KEY)
        sub_category = getattr(telemetry, st.SUB_CATEGORY_KEY)
        source_name = getattr(telemetry, st.SOURCE_NAME_KEY)
        process_type = getattr(telemetry, st.PROCESS_TYPE_KEY)
        start_date_time = getattr(telemetry, st.START_TIME).isoformat()
        run_time_in_seconds = getattr(telemetry, st.RUN_TIME)
        traffic_light = getattr(telemetry, st.TRAFFIC_LIGHT_KEY)
        io_time_in_seconds = getattr(telemetry, st.IO_TIME_KEY)
        telemetry_dict = {
            k: v.__dict__ for k, v in getattr(telemetry, st.TELEMETRY_FIELD_KEY).items()
        }
        telemetry_json = json.dumps(telemetry_dict)

        self.db_cursor.execute(
            "insert into telemetry values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            [
                telemetry_type,
                category,
                sub_category,
                source_name,
                process_type,
                start_date_time,
                str(run_time_in_seconds),
                telemetry_json,
                traffic_light,
                str(io_time_in_seconds),
            ],
        )

    def select_records(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator:
        """
        Select telemetry records unique to a single process, source category
        and sub category for as specific time period.
        """
        if not self.db_cursor:
            raise exceptions.StorageNotInitialized

        select_statement = (
            "SELECT * FROM telemetry WHERE "
            f"telemetry_type='{telemetry_type}' AND "
            f"category='{category}' AND "
            f"sub_category='{sub_category}' AND "
            f"source_name='{source_name}' AND "
            f"process_type='{process_type}' AND "
            f"start_date_time >= '{from_date_time}' AND "
            f"start_date_time < '{to_date_time}'"
        )

        return self.db_cursor.execute(select_statement)

    def _remove_existing_aggregation_telemetry(self, telemetry: TelemetryModel) -> None:
        """
        Removes any already existing aggregations for a specific telemetry
        aggregation.
        If you want to run and store a new aggregation object (for example a
        daily aggrgation) then the already daily aggregation for that day must
        be removed.

        Args:
            telemetry (TelemetryModel): The new telemetry aggregation object
        """
        query_params_exist_aggr = self._get_aggr_telem_query_params(telemetry)
        self.delete_records(**query_params_exist_aggr)

    def delete_records(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> None:
        """
        Delete telemetry records unique to a single process, source category
        and sub category for as specific time period.
        """
        if not self.db_cursor:
            raise exceptions.StorageNotInitialized

        select_statement = (
            "DELETE * FROM telemetry WHERE "
            f"telemetry_type='{telemetry_type}' AND "
            f"category='{category}' AND "
            f"sub_category='{sub_category}' AND "
            f"source_name='{source_name}' AND "
            f"process_type='{process_type}' AND "
            f"start_date_time > '{from_date_time}' AND "
            f"start_date_time < '{to_date_time}'"
        )

        self.db_cursor.execute(select_statement)



"""
module to test custom exceptions in pipeline telemetry module
"""

import pytest
from test_data import TEST_PROCESS_TYPE

from pipeline_telemetry.settings import exceptions


def raise_exception(exception, value):
    """test helper method to raise exception"""
    if value:
        raise exception(value)
    raise exception


def test_field_name_mandatory_exception():
    """test FieldNameMandatory exception"""
    with pytest.raises(exceptions.FieldNameMandatory) as exception:
        raise_exception(exceptions.FieldNameMandatory, "test_field_name")

    assert "For `test_field_name` instruction" in str(exception)


def test_unknown_instruction_exception():
    """test UnknownInstruction exception"""
    with pytest.raises(exceptions.UnknownInstruction) as exception:
        raise_exception(exceptions.UnknownInstruction, "test_instruction")

    assert "test_instruction not registered" in str(exception)


def test_instruction_registered_twice_exception():
    """test InstructionRegisteredTwice exception"""

    class TestInstructrion:
        """test class"""

        INSTRUCTION = "test_instruction"

    with pytest.raises(exceptions.InstructionRegisteredTwice) as exception:
        raise_exception(exceptions.InstructionRegisteredTwice, TestInstructrion)

    assert "test_instruction from class TestInstructrion." in str(exception)


def test_rule_can_only_have_one_instruction_exception():
    """test RuleCanHaveOnlyOneInstruction exception"""
    with pytest.raises(exceptions.RuleCanHaveOnlyOneInstruction) as exception:
        raise_exception(
            exceptions.RuleCanHaveOnlyOneInstruction, {"key1": 1, "key2": 2}
        )

    assert "Rule contains multiple keys" in str(exception)
    assert "key1, key2" in str(exception)


def test_base_count_for_sub_process_not_added_exception():
    """test BaseCountForSubProcessNotAdded exception"""
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded) as exception:
        raise_exception(exceptions.BaseCountForSubProcessNotAdded, "sub_process")

    assert "Sub process sub_process has not yet been initialized" in str(exception)


def test_telemetry_object_already_closed_exception():
    """test TelemetryObjectAlreadyClosed exception"""
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed) as exception:
        raise_exception(exceptions.TelemetryObjectAlreadyClosed, None)

    assert "Telemetry object is already closed" in str(exception)


def test_storage_class_of_incorrect_type_exception():
    """test StorageClassOfIncorrectType exception"""
    with pytest.raises(exceptions.StorageClassOfIncorrectType) as exception:
        raise_exception(exceptions.StorageClassOfIncorrectType, "Incorrect_class")

    assert "StorageClass `Incorrect_class` not a child class" in str(exception)


def test_expected_count_must_be_positive_int_exception():
    """test ExpectedCountMustBePositiveInt exception"""
    with pytest.raises(exceptions.ExpectedCountMustBePositiveInt) as exception:
        raise exceptions.ExpectedCountMustBePositiveInt

    assert "Ruleset field `expected_count` can not be negative" in str(exception)


def test_invalid_sub_process():
    """test InvalidSubProcess exception"""
    with pytest.raises(exceptions.InvalidSubProcess) as exception:
        raise exceptions.InvalidSubProcess("test_process", TEST_PROCESS_TYPE)

    assert "Sub Process `test_process` does not exist" in str(exception)


def test_process_type_not_registered():
    """test ProcessTypeNotRegistered exception"""
    with pytest.raises(exceptions.ProcessTypeNotRegistered) as exception:
        raise exceptions.ProcessTypeNotRegistered(TEST_PROCESS_TYPE)

    assert "provided process_type test_process_type not registered" in str(exception)


def test_process_types_not_of_right_class():
    """test ProcessTypesMustBeOfClassBaseEnumertor exception"""
    with pytest.raises(exceptions.ProcessTypesMustBeOfClassBaseEnumertor) as exception:
        raise exceptions.ProcessTypesMustBeOfClassBaseEnumertor

    assert "provided process_types enumerator not of class BaseEnumrator" in str(
        exception
    )


def test_sub_process_not_defined_in_process_type():
    """test InvalidSubProcessForProcessType exception"""
    with pytest.raises(exceptions.InvalidSubProcessForProcessType) as exception:
        raise exceptions.InvalidSubProcessForProcessType

    assert "Provided sub process is not defined in ProcessType." in str(exception)


def test_invalid_telemetry_type():
    """test InvalidTelemetryType exception"""
    with pytest.raises(exceptions.InvalidTelemetryType) as exception:
        raise exceptions.InvalidTelemetryType(["type1", "type2"])

    assert "Telemetry Type must be of type: type1" in str(exception)


def test_class_telemetry_params_not_defined():
    """test ClassTelemetryParamsNotDefined exception"""
    obj = "object"
    with pytest.raises(exceptions.ClassTelemetryParamsNotDefined) as exception:
        raise exceptions.ClassTelemetryParamsNotDefined(obj)

    assert f"Telemetry params not defined for class {obj.__class__.__name__}" in str(
        exception
    )


def test_unknown_telemety_type():
    """test UnknownTelemetryType exception"""
    telem_type = "SINGLE"
    with pytest.raises(exceptions.UnknownTelemetryType) as exception:
        raise exceptions.UnknownTelemetryType(telem_type)

    assert f"Unknown telemetry_type `{telem_type}` used in aggregator" in str(exception)


def test_must_have_key_mandatory():
    """test MustHaveKeyMandatory exception"""
    instruction = "instruction"
    with pytest.raises(exceptions.MustHaveKeyMandatory) as exception:
        raise exceptions.MustHaveKeyMandatory(instruction)

    assert f"For `{instruction}` instruction" in str(exception)


def test_requested_date_time_range_method_not_found():
    """test RequestedDataTimeRangeMethodNotFound exception"""
    telemetry_aggr_type = "telemetry_aggr_type"
    with pytest.raises(exceptions.RequestedDataTimeRangeMethodNotFound) as exception:
        raise exceptions.RequestedDataTimeRangeMethodNotFound(telemetry_aggr_type)

    assert f"No date_time_range method found for {telemetry_aggr_type}." in str(
        exception
    )



"""
Module to define TelemetryMixin class for adding methods to
a class that allow easy Telemetry updates
"""

from typing import Any, List, Union

from errors import ReturnValueWithStatus

from .helper import (
    add_errors_from_return_value,
    add_telemetry_counters_from_return_value,
)
from .main import Telemetry
from .settings.data_class import TelemetryCounter


class TelemetryMixin:
    _telemetry: Telemetry

    def process_errors_from_return_value(
        self, sub_process: str, return_value: ReturnValueWithStatus
    ) -> None:
        """
        Adds errors from return_value to _telemetry object for a specific
        sub_process.

        Args:
            sub_process (str):
                One of the sub_processed defined in process type for this
                telemetry object.
            return_value (ReturnValueWithStatus):
                Return value object containing the errors
        """
        add_errors_from_return_value(
            object_with_telemetry=self,
            sub_process=sub_process,
            return_value=return_value,
        )

    def process_telemetry_counters_from_return_value(
        self, return_value: ReturnValueWithStatus
    ) -> List[Any]:
        """
        Adds TelemetryCounters from return_value to _telemetry object for a
        specific sub_process.

        Args:
            return_value (ReturnValueWithStatus):
                Return value object containing the errors

        Returns:
            List:
                List with all value from return_value.result that are not of
                type TelemetryCounter. I.e. all the result entries that are not
                processed by this method.
        """
        return add_telemetry_counters_from_return_value(
            object_with_telemetry=self, return_value=return_value
        )

    def process_telemetry_counters_from_list(
        self, result_list: List[Union[Any, TelemetryCounter]]
    ) -> list[Any]:
        """
        Adds TelemetryCounters from result_list to _telemetry object.

        Args:
            result_list ([Any, TelemetryCounter]):
                Result list containing data objects (not processed) and
                Telemetry counters

        Returns:
            List:
                List with all data objects from result_list. All objects of type
                TelemetryCounter will have been removed.
        """
        return_value = ReturnValueWithStatus(result=result_list)
        return self.process_telemetry_counters_from_return_value(return_value)

    def set_telemetry_source_name(self, source_name: str) -> None:
        """(re)Set the telemerty source name

        Args:
            source_name (str): The source_name that should be added to the
                               telemetry object.
        """
        self._telemetry.telemetry.source_name = source_name



"""Module to test the DailyAggregator class."""

from test_aggregator_data import (
    DAY_BEFORE_YESTERDAY,
    TODAY,
    YESTERDAY,
    TelemetryTestList,
)

from pipeline_telemetry import DailyAggregator, TelemetrySelector
from pipeline_telemetry.aggregator.helper import TelemetryListArgs
from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings import settings as st
from pipeline_telemetry.settings.date_ranges import get_daily_date_range_yesterday

keys_in_telemetry_list_params = [
    "telemetry_type",
    "category",
    "sub_category",
    "source_name",
    "process_type",
    "from_date_time",
    "to_date_time",
]


def test_daily_aggregator_exists():
    """Test that the DailyAggregator class exists."""
    assert DailyAggregator


def test_telemetry_selector_class_exists():
    """Test that the TelemetrySelector class exists."""
    assert TelemetrySelector


def test_telemetry_selector_class_can_be_instantiated():
    """Test that the TelemetrySelector class exists."""
    assert TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )


def test_daily_aggregator_defines_target_telemetry_object():
    """Test that the DailyAggregator defines ."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )

    aggr = DailyAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=TelemetryTestList
    )
    assert aggr.target_telemetry.telemetry_type == st.DAILY_AGGR_TELEMETRY_TYPE


def test_telememtry_list_params():
    """Test if telemetry list params returns TelemetryListArgs."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )

    aggr = DailyAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=TelemetryTestList
    )
    telemetry_list_params = aggr._telememtry_list_params(
        date_time_range=next(get_daily_date_range_yesterday())
    )
    assert isinstance(telemetry_list_params, TelemetryListArgs)


def test_run_aggregation():
    """Method to test a single aggregation."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )
    aggr = DailyAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=TelemetryTestList()
    )
    result_telemetry = aggr._run_aggregation(next(get_daily_date_range_yesterday()))
    assert isinstance(result_telemetry, TelemetryModel)
    assert result_telemetry.telemetry["DATA_STORAGE"].base_counter == 3
    assert result_telemetry.telemetry["DATA_STORAGE"].fail_counter == 1


def test_run_aggregation_for_one_day():
    """Method to test that one aggregation was stored."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )
    storage = TelemetryTestList()
    aggr = DailyAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=storage
    )
    aggr.aggregate(YESTERDAY, TODAY)
    assert len(storage.stored_telemetry) == 1
    result_telemetry = storage.stored_telemetry[0]
    assert result_telemetry.telemetry["DATA_STORAGE"].base_counter == 3


def test_run_aggregation_for_two_days():
    """Method to test that two aggregations were stored."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )
    storage = TelemetryTestList()
    aggr = DailyAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=storage
    )
    aggr.aggregate(DAY_BEFORE_YESTERDAY, TODAY)
    assert len(storage.stored_telemetry) == 2
    first_telemetry = storage.stored_telemetry[0]
    second_telemetry = storage.stored_telemetry[1]
    assert first_telemetry.start_date_time.date() == DAY_BEFORE_YESTERDAY
    assert second_telemetry.start_date_time.date() == YESTERDAY



"""[summary]"""

from dataclasses import dataclass, field
from typing import Any, List, Optional

from errors import ErrorCode

from pipeline_telemetry.settings import exceptions


@dataclass(frozen=True)
class ProcessType:
    """Immutable dataclass to define a process type and its subtypes."""

    process_type: str
    subtypes: list = field(default_factory=lambda: [])

    @property
    def name(self) -> str:
        """Property method to return process key name."""
        return self.process_type

    @property
    def sub_processes(self) -> List[str]:
        """Property method to return list of sub process types."""
        return self.subtypes


@dataclass(frozen=True)
class TelemetryCounter:
    """Immutable dataclass to define a process type and its subtypes."""

    sub_process: str
    process_types: Optional[List[ProcessType]] = None
    process_type: Optional[ProcessType] = None
    counter_name: Optional[str] = None
    increment: int = 1
    error: Optional[ErrorCode] = None

    def __hash__(self):
        hash_list = [
            process_type.process_type for process_type in self.process_types or []
        ]
        hash_list.append(self.sub_process)
        if self.process_type:
            hash_list.append(self.process_type.process_type)
        if self.counter_name:
            hash_list.append(self.counter_name)
        if self.error:
            hash_list.append(self.error.code)
        hash_list.append(str(self.increment))
        return hash(tuple(hash_list))

    def add_to(
        self, object_with_telemetry: Any, increment: Optional[int] = None
    ) -> None:
        """
        Method to add to add self (the TelemetryCounter) to an object telemetry
        instance
        """
        object_with_telemetry._telemetry.add_telemetry_counter(
            telemetry_counter=self, increment=increment
        )

    def validate_sub_process(self) -> None:
        """
        Raises exception if sub_process not define in ProcessType in scope.
        """
        subtypes = []
        for process_type in self.all_process_types:
            subtypes.extend(process_type.subtypes)

        if self.sub_process not in subtypes:
            raise exceptions.InvalidSubProcessForProcessType

    @property
    def all_process_types(self) -> List[ProcessType]:
        """Returns all process_types in scope in a list

        Returns:
            List[ProcessType]: list with process_types
        """
        all_process_types = [self.process_type]
        if self.process_types:
            all_process_types.extend(self.process_types)

        return [process_type for process_type in all_process_types if process_type]

    def set_increment(self, increment: int) -> "TelemetryCounter":
        """Returns same telemetry counter with a new increment."""
        return TelemetryCounter(
            sub_process=self.sub_process,
            process_types=self.process_types,
            process_type=self.process_type,
            counter_name=self.counter_name,
            increment=increment,
            error=self.error,
        )



"""Module to define HasKey class validator"""

import jmespath
from errors import ErrorCode, ListErrors

from .abstract_validator_instruction import (
    AbstractValidatorInstruction,
    BaseValidatorInstructionRuleData,
)


class HasKey(AbstractValidatorInstruction):
    """
    class to define HasFieldName validation instruction.
    """

    INSTRUCTION = "has_key"
    FIELDNAME = "field_name"

    @classmethod
    def _validate(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - error in case
        """
        fieldname = cls._get_field_name(rule_data)
        if not jmespath.search(fieldname, dict_to_validate):
            return cls._validation_error(ListErrors.KEY_NOT_FOUND, fieldname)

        return []

    @staticmethod
    def _validation_error(error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """returns error code object in list with fieldname as error data"""
        key_specific_error = error_code.code + "@KEY_<" + fieldname + ">"
        return [ErrorCode(code=key_specific_error, description=error_code.description)]



"""
Module to test telemetry main class for pipeline telemetry module.
"""

from datetime import datetime

import pytest
from test_data import DEFAULT_TELEMETRY_PARAMS, TEST_PROCESS_TYPE_3

from pipeline_telemetry import Telemetry
from pipeline_telemetry.data_classes import TelemetryData, TelemetryModel
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings import settings as st
from pipeline_telemetry.settings.data_class import ProcessType
from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage

# pylint: disable=protected-access


def test_telemetry_exists():
    """check that Telemetry class exists"""
    assert Telemetry


def test_telemetry_instance_creation():
    """check that Telemetry instance can be created using default_params"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS)


def test_telemetry_instance_has_telemetry_type_property():
    """check that Telemetry instance has a telemetry_type property"""
    assert (
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry_type
        == st.DEFAULT_TELEMETRY_TYPE
    )


def test_telemetry_instance_has_source_name_property():
    """check that Telemetry instance has a source_name property"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).source_name == "load_weather_data"


def test_telemetry_instance_has_category_property():
    """check that Telemetry instance has a source_name property"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).category == "WEATHER"


def test_telemetry_instance_has_sub_category_property():
    """check that Telemetry instance has a sub_category property"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_category == "DAILY_PREDICTIONS"


def test_telemetry_instance_has_traffic_light_property():
    """
    Test that Telemetry instance has a traffic_light property that is
    set to the default value."""
    assert (
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).traffic_light
        == st.DEFAULT_TRAFIC_LIGHT_COLOR
    )


def test_telemetry_instance_has_sub_process_types_property():
    """check that Telemetry instance has a sub_processes_types property"""
    assert (
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_process_types
        == DEFAULT_TELEMETRY_PARAMS["process_type"].sub_processes
    )


def test_telemetry_instance_has_telemetry_property():
    """
    Test that a Telemetry instance has a telemetry property that is an
    instance of class TelemetryModel"""
    assert isinstance(Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry, TelemetryModel)


def test_telemetry_instance_has_a_start_date_time_property():
    """
    Test that a Telemetry instance has a start_date_time property that is an
    instance of class TelemetryModel"""
    assert isinstance(Telemetry(**DEFAULT_TELEMETRY_PARAMS).start_date_time, datetime)


def test_telemetry_instance_has_io_time_in_seconds_property():
    """check that Telemetry instance has a io_time_in_seconds property"""
    assert hasattr(Telemetry(**DEFAULT_TELEMETRY_PARAMS), st.IO_TIME_KEY)


def test_telemetry_instance_creation_raises_excption():
    """
    Check that Telemetry instance creation raises exception with non registered
    ProcessType.
    """
    process_type = ProcessType(process_type="not_registered", subtypes=["test"])
    telemetry_params = {
        "category": "WEATHER",
        "sub_category": "DAILY_PREDICTIONS",
        "source_name": "load_weather_data",
        "process_type": process_type,
    }
    with pytest.raises(exceptions.ProcessTypeNotRegistered):
        Telemetry(**telemetry_params)


def test_telemetry_instance_raises_exception_with_invalid_process_type():
    """
    check that Telemetry instance creation raises exception when invalid
    process_type is used
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"process_type": "invalid"}
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry(**telemetry_params)


def test_telemetry_instance_sets_io_time_to_zero():
    """
    check that Telemetry instance initially has zero io_time.
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert getattr(telemetry, st.IO_TIME_KEY) == 0


def test_telemetry_instance_increase_io_time():
    """
    check increase_io_time method increase the io_time of a Telemetry instance
    with given float value.
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry.increase_io_time(1.1)
    assert getattr(telemetry, st.IO_TIME_KEY) == 1.1


def test_initialize_new_sub_process_in_telemetry_instance():
    """
    Test that a new sub_process can be initialized.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")
    assert "RETRIEVE_RAW_DATA" in telemetry_inst.telemetry_data


def test_initialize_a_sub_process_twice_raises_an_exception():
    """
    Test that a new sub_process can be initialized only once.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")
    with pytest.raises(exceptions.SubProcessAlreadyInitialized):
        telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")


def test_initializing_non_existing_sub_process_raises_exception():
    """
    Check that initializing a non existing sub_process raises an excpetion.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.InvalidSubProcess):
        telemetry_inst._initialize_sub_process("Non existing")


def test_get_telemetry_data_sub_process():
    """
    Check get method returns a TelemetryData instance for an initialized.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")
    assert isinstance(telemetry_inst.get("RETRIEVE_RAW_DATA"), TelemetryData)


def test_increase_sub_process_base_count():
    """
    Check that a sub process base count can be increased.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_inst.get("RETRIEVE_RAW_DATA"), st.BASE_COUNT_KEY) == 1


def test_increase_sub_process_fail_count():
    """
    Check that a sub process base count can be increased.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_fail_count("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_inst.get("RETRIEVE_RAW_DATA"), st.FAIL_COUNT_KEY) == 1


def test_increase_non_exsiting_sub_process_raises_exception():
    """
    Check that a add a based count for a non existing sub process raises
    an excpetion
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.InvalidSubProcess):
        telemetry_inst.increase_sub_process_base_count("Non exsiting")


def test_increase_sub_process_base_count_to_telemetry():
    """
    Check that an existing sub process base count can be inreased multiple
    times and with different values
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA", increment=2)
    assert getattr(telemetry.get("RETRIEVE_RAW_DATA"), st.BASE_COUNT_KEY) == 3


def test_increase_sub_process_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")


def test_can_not_close_closed_telemetry_instance():
    """
    check that a closed telemetry instance can not be closed again
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.save_and_close()


def test_increase_sub_process_fail_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_fail_count("RETRIEVE_RAW_DATA")


def test_sub_process_custom_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process custom count method raises an excpetion
    when telemetry object has been closed
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_custom_count(
            sub_process="RETRIEVE_RAW_DATA", custom_counter="test_counter"
        )


def test_add_sub_process_custom_counter_to_telemetry():
    """
    Check that a custom counter can be added to a subprocess.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_custom_count(
        custom_counter="network_error", sub_process="RETRIEVE_RAW_DATA"
    )
    tel_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert tel_data.counters["network_error"] == 1


def test_increase_existing_sub_process_custom_counter_to_telemetry():
    """
    Check that a existing custom counter can be increased
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    for _ in [1, 2]:
        telemetry_inst.increase_sub_process_custom_count(
            custom_counter="network_error", sub_process="RETRIEVE_RAW_DATA"
        )

    tel_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert tel_data.counters["network_error"] == 2


def test_increase_sub_process_custom_count_with_two():
    """
    check that increase_sub_process_custom_count method creates counter
    with initial value 2
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.increase_sub_process_custom_count(
        sub_process="RETRIEVE_RAW_DATA", custom_counter="test_counter", increment=2
    )

    tel_data = telemetry_inst.get("RETRIEVE_RAW_DATA")

    assert tel_data.counters["test_counter"] == 2


def test_close_telemetry_instance_sets_run_time(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        (
            "pipeline_telemetry.storage.memory."
            "TelemetryInMemoryStorage.store_telemetry"
        ),
        return_value=None,
    )
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_result = telemetry_inst.save_and_close()
    assert float(getattr(telemetry_result, st.RUN_TIME)) > 0


def test_close_telemetry_instance_calls_store_telemetry(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        (
            "pipeline_telemetry.storage.memory."
            "TelemetryInMemoryStorage.store_telemetry"
        ),
        return_value=None,
    )
    _store_telemetry_spy = mocker.spy(TelemetryInMemoryStorage, "store_telemetry")
    Telemetry(**DEFAULT_TELEMETRY_PARAMS).save_and_close()
    assert _store_telemetry_spy.called


def test_get_storage_class_returns_in_memory_storage_by_default():
    """
    check that _get_storage_class method returns TelemetryInMemoryStorage
    instance if no storage class is provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert issubclass(telemetry_inst._storage_class, TelemetryInMemoryStorage)


def test_storage_class_returns_the_correct_storage_class():
    """
    check that storage_class attribute returns the correct
    Storage class.
    """

    # pylint: disable=too-few-public-methods
    class TestStorage(AbstractTelemetryStorage):
        """test class"""

        def store_telemetry(self, telemetry):
            """test method"""
            return None

    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy()
    telemetry_params["storage_class"] = TestStorage
    telemetry_inst = Telemetry(**telemetry_params)
    assert telemetry_inst.storage_class is TestStorage


def test_storage_class_close_method_closes_db():
    """
    check that _get_storage_class method returns an instance of
    Storage class provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    storage_instance = telemetry_inst.storage_class()
    storage_instance.close_db()
    assert storage_instance.db_in_memory is None
    assert storage_instance.db_cursor is None


def test_new_telemetry_has_default_traffic_light_color():
    """New telemetry instance should have default trafic light color."""
    assert (
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry.traffic_light
        == st.DEFAULT_TRAFIC_LIGHT_COLOR
    )


def test_set_telemetry_traffic_light_to_orange():
    """New telemetry instance traffic light property can be set to orange."""
    tele_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    tele_inst.set_orange_traffic_light()
    assert tele_inst.telemetry.traffic_light == st.TRAFIC_LIGHT_COLOR_ORANGE


def test_set_telemetry_traffic_light_to_red():
    """New telemetry instance traffic light property can be set to red."""
    tele_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    tele_inst.set_red_traffic_light()
    assert tele_inst.telemetry.traffic_light == st.TRAFIC_LIGHT_COLOR_RED


def test_sub_process_is_initialized_returns_false():
    """
    Check that _sub_process_is_initialized returns False when sub process is not
    yet initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert not telemetry_inst._sub_process_is_initialized("RETRIEVE_RAW_DATA")


def test_sub_process_is_initialized_returns_true():
    """
    Check that _sub_process_is_initialized returns true when sub process is
    initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")
    assert telemetry_inst._sub_process_is_initialized("RETRIEVE_RAW_DATA")


def test_sub_process_not_yet_initialized_returns_false():
    """
    Check that _sub_process_not_yet_initialized returns False when sub process
    initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst._initialize_sub_process("RETRIEVE_RAW_DATA")
    assert not telemetry_inst._sub_process_not_yet_initialized("RETRIEVE_RAW_DATA")


def test_sub_process_not_yet_initialized_returns_true():
    """
    Check that _sub_process_not_yet_initialized returns false when sub process
    is not yet initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert telemetry_inst._sub_process_not_yet_initialized("RETRIEVE_RAW_DATA")


def test_instanciating_telemetry_with_invalid_telemetry_type():
    """
    Test that instanciating a telemetry object with invalid telemetry type
    raises an exception.
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy()
    telemetry_params["telemetry_type"] = "invalid_type"
    with pytest.raises(exceptions.InvalidTelemetryType):
        Telemetry(**telemetry_params)


def test_validate_process_type_returns_none_with_valid_process_type():
    """
    Test that _valid_process_type validates a valid process_type by returning
    None.
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert telemetry._validate_process_type() is None


def test_validate_process_raises_exception_process_type_of_invalid_type():
    """
    Test that _valid_process_type raises exception when process_type is not
    of type ProcessType.
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy()
    telemetry_params["process_type"] = "process_type_of_invalid_type"
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry(**telemetry_params)


def test_validate_process_raises_exception_if_process_type_not_registered():
    """
    Test that _valid_process_type raises exception when process_type is not
    registered.
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy()
    telemetry_params["process_type"] = TEST_PROCESS_TYPE_3
    with pytest.raises(exceptions.ProcessTypeNotRegistered):
        Telemetry(**telemetry_params)



"""[summary]"""

from enum import Enum
from typing import Callable, Dict, List

import pipeline_telemetry.settings.date_ranges as dr
from pipeline_telemetry.settings.data_class import ProcessType

SINGLE_TELEMETRY_TYPE = "SINGLE TELEMETRY"
DAILY_AGGR_TELEMETRY_TYPE = "DAILY AGGREGATION"
WEEKLY_AGGR_TELEMETRY_TYPE = "WEEKLY AGGREGATION"
MONTHLY_AGGR_TELEMETRY_TYPE = "MONTHLY AGGREGATION"
QUARTERLY_AGGR_TELEMETRY_TYPE = "QUARTERLY AGGREGATION"
PARTIAL_AGGR_TELEMETRY_TYPE = "PARTIAL AGGREGATION"

# Constant to define date_range method specific to a given
# TELEMETRY aggregation type
AGGR_DATE_TIME_RANGE_METHODS: dict[str, Callable] = {
    DAILY_AGGR_TELEMETRY_TYPE: dr.get_daily_date_range_for_single_date,
    WEEKLY_AGGR_TELEMETRY_TYPE: dr.get_weekly_date_range_for_single_date,
    MONTHLY_AGGR_TELEMETRY_TYPE: dr.get_monthly_date_range_for_single_date,
}

DEFAULT_TELEMETRY_TYPE = SINGLE_TELEMETRY_TYPE
TELEMETRY_TYPES: List[str] = [
    PARTIAL_AGGR_TELEMETRY_TYPE,
    DEFAULT_TELEMETRY_TYPE,
    DAILY_AGGR_TELEMETRY_TYPE,
    WEEKLY_AGGR_TELEMETRY_TYPE,
    MONTHLY_AGGR_TELEMETRY_TYPE,
    QUARTERLY_AGGR_TELEMETRY_TYPE,
]

telemetry_types: Dict[str, str] = {
    type.upper().replace(" ", "_"): type.upper() for type in TELEMETRY_TYPES
}

TRAFIC_LIGHT_COLOR_GREEN = "GREEN"
TRAFIC_LIGHT_COLOR_ORANGE = "ORANGE"
TRAFIC_LIGHT_COLOR_RED = "RED"
DEFAULT_TRAFIC_LIGHT_COLOR: str = TRAFIC_LIGHT_COLOR_GREEN


BASE_COUNT_KEY = "base_counter"
ERRORS_KEY = "errors"
COUNTERS_KEY = "counters"
FAIL_COUNT_KEY = "fail_counter"
SOURCE_NAME_KEY = "source_name"
CATEGORY_KEY = "category"
SUB_CATEGORY_KEY = "sub_category"
PROCESS_TYPE_KEY = "process_type"
START_TIME = "start_date_time"
RUN_TIME = "run_time_in_seconds"
TRAFFIC_LIGHT_KEY = "traffic_light"
TELEMETRY_TYPE_KEY = "telemetry_type"
IO_TIME_KEY = "io_time_in_seconds"
TELEMETRY_FIELD_KEY = "telemetry"
AGGREGATION_KEY = "telemetry_aggregation_stats"

DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES = [
    "RETRIEVE_RAW_DATA",
    "DATA_CONVERSION",
    "DATA_STORAGE",
]

DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES = [
    "DATA_SELECTION",
    "DATA_CONVERSION",
    "DATA_UPLOAD",
]


class BaseEnumerator(Enum):
    """
    Class to define base enumerator with a keys method.
    Only to be used for defining new enumerators.
    """

    @classmethod
    def keys(cls):
        return list(cls.__members__.keys())


class DefaultProcessTypes(BaseEnumerator):
    """
    Class to define the defailt process types with their subtypes
    """

    CREATE_DATA_FROM_URL = ProcessType(
        process_type="create_data_from_url",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_API = ProcessType(
        process_type="create_data_from_api",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_FILE = ProcessType(
        process_type="create_data_from_file",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    UPLOAD_DATA = ProcessType(
        process_type="upload_data", subtypes=DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES
    )



"""
Module to test telemetry main class for pipeline telemetry module.
"""

from test_data import TEST_ERROR_CODE, TEST_ERROR_CODE_2

from pipeline_telemetry.data_classes import TelemetryData
from pipeline_telemetry.settings import settings as st


def test_telemetry_data_exists():
    """check that TelemetryData class exists"""
    assert TelemetryData


def test_telemetry_data_inc_base_count():
    tel_data = TelemetryData()
    assert getattr(tel_data, st.BASE_COUNT_KEY) == 0
    tel_data.increase_base_count(increment=1)
    assert getattr(tel_data, st.BASE_COUNT_KEY) == 1


def test_telemetry_data_inc_fail_count():
    tel_data = TelemetryData()
    assert getattr(tel_data, st.FAIL_COUNT_KEY) == 0
    tel_data.increase_fail_count(increment=1)
    assert getattr(tel_data, st.FAIL_COUNT_KEY) == 1


def test_telemetry_data_inc_errors_count():
    tel_data = TelemetryData()
    tel_data.increase_error_count(increment=1, error_code=TEST_ERROR_CODE)
    assert getattr(tel_data, st.ERRORS_KEY)[TEST_ERROR_CODE.code] == 1


def test_telemetry_data_inc_custom_count():
    tel_data = TelemetryData()
    tel_data.increase_custom_count(increment=1, counter="counter")
    assert getattr(tel_data, st.COUNTERS_KEY)["counter"] == 1


def test_add_telemetry_custom_count_data():
    tel_data_1 = TelemetryData()
    tel_data_2 = TelemetryData()
    tel_data_1.increase_custom_count(increment=1, counter="counter")
    tel_data_2.increase_custom_count(increment=2, counter="counter")
    tel_data_1.increase_custom_count(increment=3, counter="counter_1")
    tel_data_2.increase_custom_count(increment=4, counter="counter_2")
    tel_data_1 += tel_data_2
    assert getattr(tel_data_1, st.COUNTERS_KEY)["counter"] == 3
    assert getattr(tel_data_1, st.COUNTERS_KEY)["counter_1"] == 3
    assert getattr(tel_data_1, st.COUNTERS_KEY)["counter_2"] == 4


def test_add_telemetry_base_and_fail_count_data():
    tel_data_1 = TelemetryData()
    tel_data_2 = TelemetryData()
    tel_data_1.increase_base_count(increment=1)
    tel_data_2.increase_base_count(increment=2)
    tel_data_1.increase_fail_count(increment=3)
    tel_data_2.increase_fail_count(increment=4)
    tel_data_1 += tel_data_2
    assert getattr(tel_data_1, st.BASE_COUNT_KEY) == 3
    assert getattr(tel_data_1, st.FAIL_COUNT_KEY) == 7


def test_add_telemetry_error_count_data():
    tel_data_1 = TelemetryData()
    tel_data_2 = TelemetryData()
    tel_data_1.increase_error_count(increment=1, error_code=TEST_ERROR_CODE)
    tel_data_2.increase_error_count(increment=2, error_code=TEST_ERROR_CODE)
    tel_data_2.increase_error_count(increment=4, error_code=TEST_ERROR_CODE_2)
    tel_data_1 += tel_data_2
    assert getattr(tel_data_1, st.ERRORS_KEY)[TEST_ERROR_CODE.code] == 3
    assert getattr(tel_data_1, st.ERRORS_KEY)[TEST_ERROR_CODE_2.code] == 4



"""Module to provide a storage class for using Bunnet."""

from datetime import datetime
from typing import Annotated, Iterator, Optional, Sequence, Type

from bunnet import Document, Indexed, init_bunnet
from pymongo import MongoClient

from ..data_classes import TelemetryModel
from ..settings import settings as st
from .generic import AbstractTelemetryStorage

DEFAULT_DB_NAME = "GeoDataGardenTelemetry"
DEFAULT_DB_ALIAS = "geo_datagarden"


class TelemetryBunnetModel(Document):
    """
    Class to provice telemetry Mongo Model for persistance in MongoDB using Bunnet.
    """

    category: Annotated[str, Indexed()]
    sub_category: Annotated[str, Indexed()]
    source_name: Annotated[str, Indexed()]
    process_type: Annotated[str, Indexed()]
    start_date_time: Annotated[datetime, Indexed()]
    run_time_in_seconds: str
    traffic_light: Annotated[str, Indexed()]
    telemetry_type: str
    io_time_in_seconds: float = 0
    telemetry: Optional[dict] = None

    # meta = {
    #     "db_alias": "telemetry",
    #     "indexes": [
    #         "category",
    #         "sub_category",
    #         "source_name",
    #         ("category", "sub_category", "source_name", "process_type"),
    #         "process_type",
    #         "traffic_light",
    #         "start_date_time",
    #     ],
    # }

    def to_dict(self) -> dict:
        """
        Method to convert TelemetryMongoModel instance to a dict that can be
        used to instatiate a TelemetryModel object.
        The mongo object '_id' field needs to be removed as it is not used by
        TelemetryModel.
        """
        telemetry_dict = self.model_dump()
        telemetry_dict.pop("_id", None)
        # run_time_on_seconds is stored as str in Mongo but needs to be a float
        # when processing telemetry data
        telemetry_dict[st.RUN_TIME] = float(telemetry_dict[st.RUN_TIME])
        return telemetry_dict


class TelemetryBunnetStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in mongo storage class.
    This class can be used as storage_class argument when creating
    an instance of Telemetry.
    """

    def store_aggregated_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        self._remove_existing_aggregation_telemetry(telemetry)
        self.store_telemetry(telemetry)

    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        telemetry_mongo_kwargs = self._telemetry_model_kwargs(telemetry)
        TelemetryBunnetModel(**telemetry_mongo_kwargs).save()  # type: ignore

    def _remove_existing_aggregation_telemetry(self, telemetry: TelemetryModel) -> None:
        """
        Removes any already existing aggregations for a specific telemetry
        aggregation.
        If you want to run and store a new aggregation object (for example a
        daily aggrgation) then the already daily aggregation for that day must
        be removed.

        Args:
            telemetry (TelemetryModel): The new telemetry aggregation object
        """
        query_params_exist_aggr = self._get_aggr_telem_query_params(telemetry)
        for telemetry_inst in self.select_records(**query_params_exist_aggr):
            telemetry_inst.delete()

    @staticmethod
    def _telemetry_model_kwargs(telemetry: TelemetryModel) -> dict:
        """
        Returns a dicts with kwargs that can be used to create a new
        TelemetryMongoStorage instance.
        """
        telemetry_type = getattr(telemetry, st.TELEMETRY_TYPE_KEY)
        category = getattr(telemetry, st.CATEGORY_KEY)
        sub_category = getattr(telemetry, st.SUB_CATEGORY_KEY)
        source_name = getattr(telemetry, st.SOURCE_NAME_KEY)
        process_type = getattr(telemetry, st.PROCESS_TYPE_KEY)
        start_date_time = getattr(telemetry, st.START_TIME)
        run_time_in_seconds = getattr(telemetry, st.RUN_TIME)
        traffic_light = getattr(telemetry, st.TRAFFIC_LIGHT_KEY)
        io_time_in_seconds = getattr(telemetry, st.IO_TIME_KEY)
        telemetry_data = {
            k: v.__dict__ for k, v in getattr(telemetry, st.TELEMETRY_FIELD_KEY).items()
        }

        return {
            st.TELEMETRY_TYPE_KEY: telemetry_type,
            st.CATEGORY_KEY: category,
            st.SUB_CATEGORY_KEY: sub_category,
            st.SOURCE_NAME_KEY: source_name,
            st.PROCESS_TYPE_KEY: process_type,
            st.START_TIME: start_date_time,
            st.RUN_TIME: str(round(run_time_in_seconds, 2)),
            st.TRAFFIC_LIGHT_KEY: traffic_light,
            st.TELEMETRY_FIELD_KEY: telemetry_data,
            st.IO_TIME_KEY: io_time_in_seconds,
        }

    def select_records(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator:
        """
        Select telemetry records unique to a single process, source category
        and sub category for as specific time period.
        """
        query_details = {
            "telemetry_type": telemetry_type,
            "category": category,
            "sub_category": sub_category,
            "source_name": source_name,
            "process_type": process_type,
        }

        return TelemetryBunnetModel.objects(
            start_date_time__gte=from_date_time,
            start_date_time__lt=to_date_time,
            **query_details,
        )

    @staticmethod
    def _db_object_to_dict(db_object: Document) -> dict:
        """Returns a db object as a dict object."""
        # Mongo onbject need to be converted to Dict object first
        return db_object.model_dump()


ALL_MODEL_CLASSES: Sequence[Type[Document]] = [TelemetryBunnetModel]


def init_database(
    db_url: str = "mongodb://localhost:27017",
    db: Optional[str] = DEFAULT_DB_NAME,
    alias: str = DEFAULT_DB_ALIAS,
    document_models: Optional[Sequence[Type[Document]]] = None,
    **kwargs,
):
    """
    Allows users to configure the database connection.

    Parameters:
    - alias (str): The connection alias.
    - kwargs: Additional keyword arguments supported by mongoengine.connect().
    """
    # Bunnet uses Pymongo client under the hood
    client: MongoClient = MongoClient(db_url)
    # Initialize bunnet with the Product document class
    init_bunnet(
        database=client.db_name,
        document_models=document_models or ALL_MODEL_CLASSES,  # type: ignore
    )



"""Module to define abstract storage class"""

from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import Any, Dict, Iterator, TypedDict

from ..data_classes import TelemetryData, TelemetryModel
from ..settings import (
    AGGR_DATE_TIME_RANGE_METHODS,
    CATEGORY_KEY,
    PROCESS_TYPE_KEY,
    SOURCE_NAME_KEY,
    START_TIME,
    SUB_CATEGORY_KEY,
    TELEMETRY_TYPE_KEY,
)
from ..settings.date_ranges import DateTimeRange
from ..settings.exceptions import RequestedDataTimeRangeMethodNotFound


class UniqueAggregatedTelemetryKeys(TypedDict):
    """
    Class to define what query params are needed to define a unique
    aggregated telemetry object. For a set of query params given only 0 or 1
    telemetry object is allowed to exist (i.e. a daily aggregation can only
    occur once for each day, a weekly aggregation only once for each weak etc.)
    """

    telemetry_type: str
    category: str
    sub_category: str
    source_name: str
    process_type: str
    from_date_time: datetime
    to_date_time: datetime


class AbstractTelemetryStorage(metaclass=ABCMeta):
    """Abstract Telemetry Storage class

    implements store_telemetry method that persists a given telemetry object

    Any class that stores the telemetry objects should be subclassed from
    this Abstract Class
    """

    __TELEMETRY_KEY: str = "telemetry"

    @abstractmethod
    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""

    @abstractmethod
    def select_records(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator:
        """
        Select telemetry records unique to a single process and source for as specific time period.
        """

    @abstractmethod
    def _remove_existing_aggregation_telemetry(self, telemetry: TelemetryModel) -> None:
        """
        Removes any already existing aggregations for a specific telemetry
        aggregation.
        If you want to run and store a new aggregation object (for example a
        daily aggrgation) then the already daily aggregation for that day must
        be removed.

        Args:
            telemetry (TelemetryModel): The new telemetry aggregation object
        """

    def _telemetry_storage_to_object(
        self, stored_telemetry_object: Dict
    ) -> TelemetryModel:
        """Method to convert a sql light object into"""
        telemetry_data = stored_telemetry_object.pop(self.__TELEMETRY_KEY, "{}")
        telemetry_model = TelemetryModel(**stored_telemetry_object)
        for sub_process, sub_process_tel_data in telemetry_data.items():
            telemetry_model.telemetry.update(
                {sub_process: TelemetryData(**sub_process_tel_data)}
            )
        return telemetry_model

    @staticmethod
    def _db_object_to_dict(db_object: Any) -> Dict:
        """Returns a db object as a dict object."""
        # If the persistance model does not return a proper Dict object then
        # override this method to ensure a Dict object is available.
        return db_object

    def _get_aggr_telem_query_params(
        self, telemetry: TelemetryModel
    ) -> UniqueAggregatedTelemetryKeys:
        """
        Returns a dict with the query params to retrieve a unique aggregated
        telemetry object.

        Args:
            telemetry (TelemetryModel): The new telemetry aggregation object

        returns:
            UniqueAggregatedTelemetryKeys:
                TypedDict with query params that can be used as input for a
                DB query to retrieve all instances of unique aggregated
                telemetry object (should in theory always return 0 or 1
                instance.
        """
        date_time_range = self._get_aggr_telem_date_time_range(telemetry)
        return UniqueAggregatedTelemetryKeys(
            telemetry_type=getattr(telemetry, TELEMETRY_TYPE_KEY),
            source_name=getattr(telemetry, SOURCE_NAME_KEY),
            category=getattr(telemetry, CATEGORY_KEY),
            sub_category=getattr(telemetry, SUB_CATEGORY_KEY),
            process_type=getattr(telemetry, PROCESS_TYPE_KEY),
            from_date_time=date_time_range.from_date,
            to_date_time=date_time_range.to_date,
        )

    @staticmethod
    def _get_aggr_telem_date_time_range(telemetry: TelemetryModel) -> DateTimeRange:
        """
        Method to get the correct datatime range for a date depending on the
        aggregation telemetry_type.
        For example for a monthly aggregation the input date of 2022-10-10
        will be converted into a datetime range of (2022-10-01, 2022-11-01).
        For a daily aggegration this would result in (2022-10-10, 2022-10-11).
        etc.
        """
        start_date_time = getattr(telemetry, START_TIME)
        telemetry_type = getattr(telemetry, TELEMETRY_TYPE_KEY)
        date_time_range_method = AGGR_DATE_TIME_RANGE_METHODS.get(telemetry_type)
        if not date_time_range_method:
            raise RequestedDataTimeRangeMethodNotFound(telemetry_type)

        return date_time_range_method(start_date_time.date())

    def telemetry_list(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator[TelemetryModel]:
        """
        Method to return an iteraror TelemetryModel instances retrieved
        from a database query with the provided arguments.
        """
        selected_records = self.select_records(
            telemetry_type=telemetry_type,
            category=category,
            sub_category=sub_category,
            source_name=source_name,
            process_type=process_type,
            from_date_time=from_date_time,
            to_date_time=to_date_time,
        )
        for record in selected_records:
            record_dict = self._db_object_to_dict(record)
            yield self._telemetry_storage_to_object(record_dict)



"""
This module provides test data for the telemetry storage tests.
"""

from test_data import DEFAULT_TELEMETRY_MODEL_PARAMS  # noqa: F401
from test_data import DEFAULT_TELEMETRY_PARAMS  # noqa: F401



from pipeline_telemetry import ProcessType, ProcessTypes, ProcessTypesMeta

FLOW_1 = [
    "GET_DATA_RESPONSES",
    "STORE_DATA",
]

FLOW_2 = [
    "GET_DATA_RESPONSES",
    "CONVERT_DATA",
]


class TestTypesOne:
    """
    Class to define the process types with their subtypes for Wearther module
    """

    FLOW_ONE = ProcessType(process_type="F1", subtypes=FLOW_1)


class TestTypesTwo:
    """
    Class to define the process types with their subtypes for Wearther module
    """

    FLOW_TWO = ProcessType(process_type="F2", subtypes=FLOW_2)

    NO_PROCESS_TYPE: str = "other attribute"


def test_process_type_meta_exists():
    assert ProcessTypesMeta


def test_meta_class_adds_process_types_to_registered_list():
    class NewClass(TestTypesOne, TestTypesTwo, metaclass=ProcessTypesMeta): ...  # noqa: E701

    # assert NewClass.is_registered(TestTypesOne.FLOW_ONE)
    assert NewClass.is_registered(TestTypesTwo.FLOW_TWO)


def test_meta_class_skips_non_process_types_attributes():
    class NewClass(TestTypesOne, TestTypesTwo, metaclass=ProcessTypesMeta): ...  # noqa: E701

    assert hasattr(TestTypesTwo, "NO_PROCESS_TYPE")
    assert not NewClass.is_registered(TestTypesTwo.NO_PROCESS_TYPE)  # type: ignore


def test_class_from_metaclass_issubclass_process_types_class():
    class NewClass(TestTypesOne, TestTypesTwo, metaclass=ProcessTypesMeta): ...  # noqa: E701

    assert issubclass(NewClass, ProcessTypes)



"""
module to test has_key instruction class
"""

# pylint: disable=protected-access
import pytest
from test_data import TEST_ERROR_CODE

from pipeline_telemetry.settings.exceptions import ExpectedCountMustBePositiveInt
from pipeline_telemetry.settings.telemetry_errors import ErrorCode
from pipeline_telemetry.validators.validate_entries import (
    ValidateEntries,
    ValidateEntriesRuleData,
)


def test_validate_entries_class_exists():
    """check that ValidateEntries class exists"""
    assert ValidateEntries


def test_validate_entries_class_is_singleton_clss():
    """Check that ValidateEntries class is a singleton class."""
    assert ValidateEntries() is ValidateEntries()


def test_validate_rule_method_with_neg_expected_count_raises_exception():
    """
    check that _validate_rule method raises exception when expected_count
    is a negative integer
    """
    with pytest.raises(ExpectedCountMustBePositiveInt):
        ValidateEntriesRuleData(**{"field_name": "test", "expected_count": -1})


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_data = ValidateEntriesRuleData(**{"field_name": "test", "expected_count": 1})
    assert ValidateEntries._get_field_name(rule_data) == "test"


def test_get_expected_count():
    """
    check that _get_expected_count method returns value of
    expected_count
    """
    rule_data = ValidateEntriesRuleData(**{"field_name": "test", "expected_count": 1})

    assert ValidateEntries._get_expected_count(rule_data) == 1


def test_validation_error_method():
    """
    check that _validation_error method returns instance
    of ErrorCode with fieldname in error_data
    """
    validation_error_list = ValidateEntries._validation_error(
        TEST_ERROR_CODE, "test_error_data"
    )
    validation_error = validation_error_list[0]
    assert isinstance(validation_error_list, list)
    assert isinstance(validation_error, ErrorCode)
    assert validation_error.error_data == "test_error_data"


def test_validate_valid_dict():
    """
    Test that a dict with a field with a list with the expected count (length)
    returns empty error list
    """
    valid_dict = {"correct_key": [1, 2, 3, 4]}
    rule_dict = {"field_name": "correct_key", "expected_count": 4}
    validation_result = ValidateEntries.validate(
        dict_to_validate=valid_dict, rule_dict=rule_dict
    )
    assert validation_result == []


def test_validate_valid_dict_with_nested_key():
    """
    Test that a dict of the right length in nested key returns an empty error
    lits
    """
    valid_dict = {"key": {"nested_key": {"a": 1, "b": 2}}}
    rule_dict = {"field_name": "key.nested_key", "expected_count": 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=valid_dict, rule_dict=rule_dict
    )
    assert validation_result == []


def test_validate_invalid_dict():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_001 and the field_name in the error_data

    """
    invalid_dict = {"incorrect_key": "value"}
    rule_dict = {"field_name": "correct_key", "expected_count": 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_dict=rule_dict
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == "VALIDATE_ENTRIES_ERR_001"
    assert validation_result[0].error_data == "correct_key"


def test_validate_dict_with_wrong_type_in_field():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_002 and the field_name in the error_data

    """
    invalid_dict = {"correct_key": "value"}
    rule_dict = {"field_name": "correct_key", "expected_count": 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_dict=rule_dict
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == "VALIDATE_ENTRIES_ERR_002"
    assert validation_result[0].error_data == "correct_key"


def test_validate_dict_with_wrong_number_of_entries():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_003 and the field_name in the error_data

    """
    invalid_dict = {"correct_key": [1, 2, 3]}
    rule_dict = {"field_name": "correct_key", "expected_count": 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_dict=rule_dict
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == "VALIDATE_ENTRIES_ERR_003"
    assert validation_result[0].error_data == "correct_key"
